/*
** Module   : Rectangle.java
** Abstract : Rectangular area descriptor.
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 CA  20140729 Added size, difference and hashCode APIs.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

/**
 * Rectangular area descriptor.
 */
public class Rectangle
{
   /** Rectangle bottom row. */
   private final int bottom;

   /** Rectangle left column. */
   private final int left;

   /** Rectangle right column. */
   private final int right;

   /** Rectangle top row. */
   private final int top;

   /**
    * Constructor.
    * 
    * @param   top
    *          Top row.
    * @param   left
    *          Left column.
    * @param   bottom
    *          Bottom row.
    * @param   right
    *          Right column.
    */
   public Rectangle(int top, int left, int bottom, int right)
   {
      this.top = top;
      this.left = left;
      this.bottom = bottom;
      this.right = right;
   }

   /**
    * Constructor.
    * 
    * @param   topleft
    *          Top-left corner.
    * @param   size
    *          Rectangle dimension.
    */
   public Rectangle(Point topleft, Dimension size)
   {
      top = topleft.y;
      left = topleft.x;
      bottom = top + size.height - 1;
      right = left + size.width - 1;
   }

   /**
    * Copy constructor.
    * 
    * @param   rect
    *          Source rectangle.
    */
   public Rectangle(Rectangle rect)
   {
      top = rect.top();
      left = rect.left();
      bottom = rect.bottom();
      right = rect.right();
   }

   /**
    * Get bottom coordinate of this rectangle.
    * 
    * @return  bottom.
    */
   public int bottom()
   {
      return bottom;
   }

   /**
    * Returns <code>true</code> if the specified point is inside this
    * rectangle.
    * 
    * @param   x
    *          X coordinate.
    * @param   y
    *          Y coordinate.
    * 
    * @return <code>true</code> if coordinates are inside rectangle.
    */
   public boolean contains(int x, int y)
   {
      return ((x >= left) && (x <= right) && (y >= top) && (y <= bottom));
   }

   /**
    * Returns <code>true</code> if the specified point is inside this
    * rectangle.
    * 
    * @param   p
    *          Point to check.
    *          
    * @return  <code>true</code> if specified point is inside rectangle.
    */
   public boolean contains(Point p)
   {
      return contains(p.x, p.y);
   }

   /**
    * Returns <code>true</code> is specified rectangle is completely inside
    * this rectangle.
    * 
    * @param   r
    *          Rectangle to check.
    *          
    * @return  <code>true</code> if specified rectangle is inside this
    *          rectangle.
    */
   public boolean contains(Rectangle r)
   {
      return ((r.left >= left) && (r.right <= right) && (r.top >= top) && (r.bottom <= bottom));
   }

   /**
    * Check if rectangle is empty.
    * 
    * @return  <code>true</code> if rectangle is empty.
    */
   public boolean empty()
   {
      return top > bottom || right < left;
   }

   /**
    * Check if this rectangle is equal to specified one.
    * 
    * @param   rect
    *          Rectangle to check.
    * 
    * @return  <code>true</code> if rectangles are equal.
    */
   public boolean equals(Rectangle rect)
   {
      if (rect == null)
         return false;
      if (top != rect.top)
         return false;
      if (bottom != rect.bottom)
         return false;
      if (left != rect.left)
         return false;
      if (right != rect.right)
         return false;
      return true;
   }
   
   /**
    * Compute the hashcode of this rectangle.
    * 
    * @return   A hashcode uniquely describing this rectangle.
    */
   @Override
   public int hashCode()
   {
      int r = 17;
      r = 37 * r + top;
      r = 37 * r + bottom;
      r = 37 * r + left;
      r = 37 * r + right;
      
      return r;
   }

   /**
    * Calculate rectangle which is intersection between this and specified
    * rectangle.
    * 
    * @param   rect
    *          Source rectangle.
    *          
    * @return  rectangle which is an intersection between this and specified
    *          rectangle or <code>null</code> if there is no intersection.
    */
   public Rectangle intersection(Rectangle rect)
   {
      if (top > rect.bottom)
         return null;
      if (bottom < rect.top)
         return null;
      if (left > rect.right)
         return null;
      if (right < rect.left)
         return null;

      return new Rectangle((top > rect.top) ? top
                                           : rect.top,
                           (left > rect.left) ? left
                                             : rect.left,
                           (bottom < rect.bottom) ? bottom
                                                 : rect.bottom,
                           (right < rect.right) ? right
                                               : rect.right);
   }

   /**
    * Check if this and specified rectangle have intersection.
    * 
    * @param   rect
    *          Source rectangle.
    * 
    * @return  <code>true</code> if rectangles have intersection.
    */
   public boolean intersects(Rectangle rect)
   {
      if (rect.left > right)
         return false;
      if (rect.right < left)
         return false;
      if (rect.top > bottom)
         return false;
      if (rect.bottom < top)
         return false;
      return true;
   }

   /**
    * Get left coordinate of this rectangle.
    * 
    * @return  left.
    */
   public int left()
   {
      return left;
   }

   /**
    * Get right coordinate of this rectangle.
    * 
    * @return  right.
    */
   public int right()
   {
      return right;
   }

   /**
    * Get top coordinate of this rectangle.
    * 
    * @return top.
    */
   public int top()
   {
      return top;
   }

   /**
    * Create new rectangle which is translated by specified offset.
    * 
    * @param   dx
    *          Delta X.
    * @param   dy
    *          Delta Y.
    * 
    * @return  translated rectangle.
    */
   public Rectangle translate(int dx, int dy)
   {
      return new Rectangle(top + dy, left + dx, bottom + dy, right + dx);
   }

   /**
    * Create new rectangle which is translated by specified offset.
    * 
    * @param   p
    *          Offset represented as a point.
    *           
    * @return  translated rectangle.
    */
   public Rectangle translate(Point p)
   {
      return translate(p.x, p.y);
   }

   /**
    * Create new rectangle which is a union of this and specified rectangle.
    * 
    * @param   rect
    *          Source rectangle.
    * 
    * @return  union rectangle.
    */
   public Rectangle union(Rectangle rect)
   {
      return new Rectangle((top < rect.top) ? top
                                           : rect.top,
                           (left < rect.left) ? left
                                             : rect.left,
                           (bottom > rect.bottom) ? bottom
                                                 : rect.bottom,
                           (right > rect.right) ? right
                                               : rect.right);
   }

   /**
    * Create new rectangles which is the difference between this and the specified rectangle.
    * 
    * @param   rect
    *          Removed rectangle.
    * @param   baseUnit
    *          The base measurement unit.
    *
    * @return  All the rectangles which are part of this rectangle but not part of the specified 
    *          rectangle.
    */
   public List<Rectangle> difference(Rectangle rect, int baseUnit)
   {
      List<Rectangle> res = new ArrayList<>(4);

      Rectangle intersection = this.intersection(rect);
      if (intersection == null)
      {
         // nothing to remove, return this
         res.add(this);

         return res;
      }

      // at this point, is just a matter of removing the intersection
      rect = intersection;

      if (rect.top - baseUnit >= top)
      {
         // rectangle found above the intersection
         res.add(new Rectangle(top, left, rect.top - baseUnit, right));
      }
      
      if (rect.bottom + baseUnit <= bottom)
      {
         // rectangle found bellow the intersection
         res.add(new Rectangle(rect.bottom + baseUnit, left, bottom, right));
      }
      
      if (rect.left - baseUnit >= left)
      {
         // rectangle found on the left-side of the intersection
         res.add(new Rectangle(rect.top, left, rect.bottom, rect.left - baseUnit));
      }
      
      if (rect.right + baseUnit <= right)
      {
         // rectangle found on the right-side of the intersection
         res.add(new Rectangle(rect.top, rect.right + baseUnit, rect.bottom, right));
      }

      return res;
   }

   /**
    * Compute the size of this rectangle.
    * 
    * @param    baseUnit
    *           A parameter indicating the base measurument unit (which can be 1 for character 
    *           interfaces or 0.01 for graphical interfaces).
    * 
    * @return   The size of this rectangle.
    */
   public int size(double baseUnit)
   {
      // TODO: change this to support decimal values
      return (int) (bottom - top + baseUnit) * (int) (right - left + baseUnit);
   }
}
