/*
** Module   : ScreenBitmap.java
** Abstract : Provide clipping behavior for drawing.
**
** Copyright (c) 2008-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100903 Fixed javadoc.
** 004 SIY 20111204 Updated imports.
** 005 CA  20140729 Removed the matrix bitmap: instead, the activated rectangles are kept in a 
**                  list, to allow decimal coordinates and a more dynamic approach on the screen's
**                  dimensions.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

/**
 * Keep the state of the screen, for clipping purposes. 
 */
public class ScreenBitmap
{
   /**
    * Base measure unit.
    * TODO: make this configurable based on the output manager in use. 
    */
   private static final int BASE_UNIT = 1;
   
   /** Invalid coordinate constant. */
   private static final int INV_COORD = Integer.MIN_VALUE;
   
   /** The screen column number. */
   int cols;
   
   /** The screen row number. */
   int rows;
   
   /** Screen state - all activated rectangles. */
   private List<Rectangle> bitmap;
   
   /** 
     * Union rectangle of all clipping areas. Null value means the screen is 
     * not in clipping mode.
     */
   private Rectangle outerRectangle;

   /** The total number of invalid characters, used for searching. */
   private int invalidChars;
   
   /** 
    * Copy bitmap, used for searching the bottommost frame from which drawing can be started. 
    */
   private List<Rectangle> searchBitmap;
   /** 
    * Flag to track if clipping rectangles were added to this bitmap 
    * (including <code>null</code> rectangles, which reset the screen). 
    */
   private boolean hasRectangles;
   
   /**
    * Default c'tor. Used for internal purposes, when a copy of an existing
    * instance is needed. This is a no-op.
    */
   private ScreenBitmap()
   {
      // no-op
   }
   
   /**
    * Initializes the bitmap with the screen size. Performs a reset, setting 
    * all bitmap values to <code>true</code>.
    * 
    * @param   rows
    *          The number of screen rows.
    * @param   cols
    *          The number of screen columns.
    */
   public ScreenBitmap(int rows, int cols)
   {
      this.rows = rows;
      this.cols = cols;
      
      bitmap = new ArrayList<>();
      
      resetScreen();
   }
   
   /**
    * Create a copy of this screen bitmap and count the invalid chars. This
    * initialization is needed to start searching for the frame from where to
    * start drawing.
    */
   void startScreenSearch()
   {
      searchBitmap = new ArrayList<>();
      invalidChars = 0;

      int rowFrom = 0;
      int rowTo   = rows;
      int colFrom = 0;
      int colTo   = cols;
      
      if (outerRectangle != null)
      {
         rowFrom = outerRectangle.top();
         rowTo   = outerRectangle.bottom() + BASE_UNIT;
         colFrom = outerRectangle.left();
         colTo   = outerRectangle.right() + BASE_UNIT;
      }

      // copy only the rows which intersect the clipped area
      Rectangle outer = new Rectangle(rowFrom, colFrom, rowTo - BASE_UNIT, colTo - BASE_UNIT);
      for (Rectangle r : bitmap)
      {
         Rectangle intersection = r.intersection(outer);
         if (intersection != null)
         {
            searchBitmap.add(intersection);

            invalidChars += intersection.size(BASE_UNIT);
         }
      }
   }
   
   /**
    * Mark a certain area as searched and decrement the total number of 
    * clipped characters.
    * 
    * @param    rect
    *           Area to mark as searched.
    * 
    * @return   <code>true</code> if this area contains a clipped character,
    *           which was not part of an upper widget.
    */
   boolean markArea(Rectangle rect)
   {
      if (rect == null || hasCoordinatesOutsideScreen(rect))
      {
         return false;
      }

      boolean marked = false;

      // remove the specified rectangle
      List<Rectangle> work = new ArrayList<>();
      for (Rectangle r : searchBitmap)
      {
         Rectangle intersection = r.intersection(rect);
         if (intersection != null)
         {
            // keep only the rectangles outside of the marked area
            work.addAll(r.difference(intersection, BASE_UNIT));
            
            // decrement the number of invalid chars
            invalidChars -= intersection.size(BASE_UNIT);
            marked = true;
         }
         else
         {
            work.add(r);
         }
      }
      
      searchBitmap.clear();
      searchBitmap.addAll(work);

      return marked;
   }
   
   /**
    * Get the total number of available clipped characters, from the search
    * bitmap screen. 
    * 
    * @return   The total number of invalid characters.
    */
   int getInvalidChars()
   {
      return invalidChars;
   }
   
   /**
    *  Reset the screen to allow drawing. All bitmap positions are set to
    *  <code>true</code>, indicating that the location can be drawn.
    */
   void resetScreen()
   {
      resetScreen(true);
      outerRectangle = null;
      hasRectangles = false;
   }

   /**
    * Reset the screen to the specified state. 
    * 
    * @param   state
    *          <code>true</code> means that position can be drawn.
    */
   public void resetScreen(boolean state)
   {
      int rowFrom = 0;
      int rowTo   = rows;
      int colFrom = 0;
      int colTo   = cols;
      
      if (outerRectangle != null)
      {
         rowFrom = outerRectangle.top();
         rowTo   = outerRectangle.bottom() + BASE_UNIT;
         colFrom = outerRectangle.left();
         colTo   = outerRectangle.right() + BASE_UNIT;
      }

      setState(rowFrom, colFrom, rowTo - BASE_UNIT, colTo - BASE_UNIT, state);
   }
   
   private void setState(int rowFrom, int colFrom, int rowTo, int colTo, boolean state)
   {
      Rectangle outer = new Rectangle(rowFrom, colFrom, rowTo, colTo);
      List<Rectangle> work = new ArrayList<>();
      
      if (state)
      {
         // this means the entire outer rectangle needs to be part of the bitmap
         work.add(outer);
      }

      for (Rectangle r : bitmap)
      {
         Rectangle intersection = r.intersection(outer);
         
         if (intersection == null)
         {
            // just keep the rectangle, is not of concern
            work.add(r);
         }
         else
         {
            // just keep the differences, regardless how the state changes:
            // 1. if the state is on, then the outer is already added, so this will force the
            //    rectangles not to overlap
            // 2. if the state is off, the intersection really needs to be removed.
            work.addAll(r.difference(intersection, BASE_UNIT));
         }
      }
      
      bitmap.clear();
      bitmap.addAll(work);
   }
   
   /**
    * Adds an invalidated rectangle to the bitmap. Invalidated rectangles are 
    * rectangles where output is allowed. If this is the first call after 
    * the screen was reset, the entire bitmap is filled with 
    * <code>false</code> values.
    * 
    * @param   rect
    *          Invalidated rectangle, which entire region will be set to
    *          <code>true</code>. If <code>null</code>, then the entire screen
    *          area is set as drawable.
    */
   void addRectangle(Rectangle rect)
   {
      // skip negative area rectangles
      if (rect != null && hasCoordinatesOutsideScreen(rect))
      {
         return;
      }

      if (outerRectangle == null && rect != null)
      {
         resetScreen(false);
      }
      
      if (rect == null)
      {
         // check if screen already in no clipping mode
         if (outerRectangle != null)
         {
            resetScreen();
         }
      }
      else
      {
         // set the state for this rectangle to true
         setState(rect.top(), rect.left(), rect.bottom(), rect.right(), true);
         
         if (outerRectangle == null)
         {
            outerRectangle = new Rectangle(rect);
         }
         else
         {
            // expand the union of clipped areas
            outerRectangle = outerRectangle.union(rect);
         }
      }

      hasRectangles = true;      
   }
   
   /**
    * Clips the bitmap so that no output is allowed outside of the given 
    * rectangle. 
    * 
    * @param   rect
    *          clipping rectangle.
    *          If <code>null</code>, nothing changes in the bitmap
    */
   void clipRectangle(Rectangle rect)
   {
      // skip null or negative area rectangles
      if (rect == null || hasCoordinatesOutsideScreen(rect))
      {
         return;
      }

      // clip off anything outside the given rectangle
      if (rect.top() > 0)
      {
         // clip full lines on the top of the screen
         setState(0, 0, rect.top() - BASE_UNIT, cols - BASE_UNIT, false);
      }

      if (rect.left() > 0)
      {
         // clip left sides of partial lines
         setState(rect.top(), 0, rect.bottom() - BASE_UNIT, rect.left() - BASE_UNIT, false);
      }

      if (rect.right() < cols - BASE_UNIT)
      {
         // clip right sides of partial lines
         setState(rect.top(), rect.right() + BASE_UNIT, 
                  rect.bottom() - BASE_UNIT, cols - BASE_UNIT, false);
      }

      if (rect.bottom() < rows - BASE_UNIT)
      {
         // clip full lines on the bottom of the screen
         setState(rect.bottom() + BASE_UNIT, 0, rows - BASE_UNIT, cols - BASE_UNIT, false);
      }

      // adjust the outer rectangle
      if (outerRectangle == null)
      {
         outerRectangle = new Rectangle(rect);
      }
      else
      {
         // limit the outer rectangle
         outerRectangle = outerRectangle.intersection(rect);
      }
   }
   
   /**
    * This method will return <code>true</code> if no clipping info was added
    * to this bitmap (i.e. the screen is in 'reset' state, with all characters
    * being drawable).
    * 
    * @return   <code>true</code> if no clipping is in effect.
    */
   boolean isScreenReset()
   {
      return outerRectangle == null;
   }
   
   /**
    * Get outer rectangle.
    * 
    * @return   outer rectangle of all clipping areas.
    */
   Rectangle getOuterRectangle()
   {
      if (outerRectangle == null)
         return null;
      
      return new Rectangle(outerRectangle);
   }
   
   /**
    * Check if the screen was added any rectangles (including null rectangles,
    * which reset the entire screen).
    * 
    * @return   the <code>hasRectangles</code> flag.
    */
   boolean hasRectangles()
   {
      return hasRectangles;
   }
   
   /**
    * Check the bitmap if drawing is possible at that location.
    * 
    * @param    row
    *           The screen row.
    * @param    col
    *           The screen column.
    * 
    * @return   <code>true</code> if drawing is allowed.
    */
   public boolean canDraw(int row, int col)
   {
      // ignore requests outside the bitmap
      if (row >= rows || col >= cols || row < 0 || col < 0)
      {
         return false;
      }

      for (Rectangle r : bitmap)
      {
         if (r.contains(col, row))
         {
            return true;
         }
      }
      
      return false;
   }

   /**
    * Set one specified bit in bitmap.
    * 
    * @param    row
    *           Bit row.
    * @param    col
    *           Bit column.
    * @param    value
    *           New bit value.
    */
   public void setBit(int row, int col, boolean value)
   {
      // ignore requests outside the bitmap
      if (row >= rows || col >= cols || row < 0 || col < 0)
      {
         return;
      }
      
      // determine if this is part of the marked rectangles
      Rectangle point = new Rectangle(row, col, row, col);
      Rectangle rset = null;
      for (Rectangle r : bitmap)
      {
         if (r.intersects(point))
         {
            if (value)
            {
               // already set, nothing to do
               return;
            }
            
            rset = r;
            break;
         }
      }
      
      if (value)
      {
         // not set, add it
         bitmap.add(point);
      }
      else
      {
         bitmap.addAll(rset.difference(point, BASE_UNIT));
      }
   }
   
   /**
    * Check the bitmap if drawing is possible at that location.
    * 
    * @param    p
    *           The screen location.
    * 
    * @return   <code>true</code> if drawing is allowed.
    */
   boolean canDraw(Point p)
   {
      return canDraw(p.y, p.x);
   }

   /**
    * Check if a given rectangle has negative coordinates.
    * 
    * @param    rect
    *           Rectangle to check for negative coordinates.
    * @return   <code>true</code> if any of its coordinates are negative;
    *           <code>false</code>, if not
    */
   boolean hasCoordinatesOutsideScreen(Rectangle rect)
   {
      if (rect.top() < 0        || 
          rect.bottom() < 0     || 
          rect.left() < 0       || 
          rect.right() < 0      ||
          rect.top() >= rows    ||
          rect.bottom() >= rows ||
          rect.left() >= cols   ||
          rect.right() >= cols)
      {
         return true;
      }
      return false;
   }

   /**
    * Check if a given rectangle has invalid coordinates.
    * 
    * @param    rect
    *           Rectangle to check for invalid coordinates.
    * @return   <code>true</code> if any of its coordinates are invalid;
    *           <code>false</code>, if not
    */
   boolean hasInvalidCoordinates(Rectangle rect)
   {
      if (rect.top()    == INV_COORD || 
          rect.bottom() == INV_COORD || 
          rect.left()   == INV_COORD || 
          rect.right()  == INV_COORD)
      {
         return true;
      }
      return false;
   }
   
   /**
    * Check if the given location can draw data.
    * 
    * @param     rect
    *            The area of the widget which is checked for drawing.
    * 
    * @return    <code>true</code> if screen any of the bitmap locations 
    *            within the given rectangle are <code>true</code>;
    *            <code>false</code> if nothing can be drawn.
    */
   boolean drawNeeded(Rectangle rect)
   {
      if (rect == null || hasCoordinatesOutsideScreen(rect))
      {
         return false;
      }
      
      for (int y = rect.top(); y <= rect.bottom(); y++)
      {
         for (int x = rect.left(); x <= rect.right(); x++)
         {
            if (canDraw(y, x))
            {
               return true;
            }
         }
      }
      return false;
   }

   /**
    * Get a copy of this bitmap instance.
    * 
    * @return   A copy of this bitmap instance.
    */
   ScreenBitmap getCopy()
   {
      ScreenBitmap sb = new ScreenBitmap();
      
      // copy the fields
      sb.outerRectangle = null;
      if (this.outerRectangle != null)
      {
         sb.outerRectangle = new Rectangle(this.outerRectangle);
      }
      sb.rows = this.rows;
      sb.cols = this.cols;
      
      // copy the bitmap
      // by default, this is filled with 'false' values 
      sb.bitmap = new ArrayList<>();

      int rowFrom = 0;
      int rowTo   = rows;
      
      if (outerRectangle != null)
      {
         rowFrom = outerRectangle.top();
         rowTo   = outerRectangle.bottom() + BASE_UNIT;
      }
      
      // copy only the rows which intersect the clipped area
      Rectangle outer = new Rectangle(rowFrom, 0, rowTo - BASE_UNIT, cols - BASE_UNIT);
      for (Rectangle r : bitmap)
      {
         Rectangle intersection = r.intersection(outer);
         if (intersection != null)
         {
            sb.bitmap.add(intersection);
         }
      }
      
      return sb;
   }
    
   /**
    * Get the drawable horizontal fragments, to the left side of the specified
    * column.
    * <p>
    * This will return an array of drawable fragments, each fragment having
    * specified its start and end locations (both inclusive), relative to the
    * 'length' parameter.
    * 
    * @param    p
    *           The location where to start looking for drawable fragments,
    *           horizontally.
    * @param    length
    *           The length of the line to look into.
    *           
    * @return   A list of drawable fragments, where {@link Point#x} represents 
    *           the start column and {@link Point#y} represents the end column,
    *           inclusive.
    */
   List<Point> getHorizontalFragments(Point p, int length)
   {
      return getHorizontalFragments(p.y, p.x, length);
   }

   /**
    * Get the drawable horizontal fragments, to the left side of the specified
    * column.
    * <p>
    * This will return an array of drawable fragments, each fragment having
    * specified its start and end locations (both inclusive), relative to the
    * 'length' parameter.
    * 
    * @param    y
    *           The row location where to start looking for drawable fragments,
    *           horizontally.
    * @param    x
    *           The column location where to start looking for drawable 
    *           fragments, horizontally.
    * @param    length
    *           The length of the line to look into.
    *           
    * @return   A list of drawable fragments, where {@link Point#x} represents 
    *           the start column and {@link Point#y} represents the end column,
    *           inclusive.
    */
   @SuppressWarnings("unchecked")
   List<Point> getHorizontalFragments(int y, int x, int length)
   {
      if (y < 0 || x < 0 || x >= cols)
      {
         return Collections.emptyList();
      }

      List<Point> frags = null;
      
      int row = y;
      int col = x;
      int end = Math.min(cols, col + length);

      if (outerRectangle == null    || 
          (!(col > outerRectangle.right() || end < outerRectangle.left())     &&
           row >= outerRectangle.top()                                        &&
           row <= outerRectangle.bottom()))
      {
         // search only if the line might intersect the clipped area
         do
         {
            while (col < end && !canDraw(row, col))
            {
               col = col + BASE_UNIT;
            }
   
            int start = col;
            while (col < end && canDraw(row, col))
            {
               col = col + BASE_UNIT;
            }
            
            if (start != col)
            {
               if (frags == null)
               {
                  frags = new ArrayList<Point>(5);
               }
               frags.add(new Point(start - x, col - x - BASE_UNIT));
            }
         }
         while (col < end);
      }
      return frags == null ? Collections.EMPTY_LIST : frags;
   }

   /**
    * Get the drawable vertical fragments, to the down side of the specified
    * row.
    * <p>
    * This will return an array of drawable fragments, each fragment having
    * specified its start and end locations (both inclusive), relative to the
    * 'length' parameter.
    * 
    * @param    p
    *           The location where to start looking for drawable fragments,
    *           vertically.
    * @param    length
    *           The length of the line to look into.
    *           
    * @return   A list of drawable fragments, where {@link Point#x} represents 
    *           the start row and {#link Point#y} represents the end row,
    *           inclusive.
    */
   List<Point> getVerticalFragments(Point p, int length)
   {
      return getVerticalFragments(p.y, p.x, length);
   }

   /**
    * Get the drawable vertical fragments, to the down side of the specified
    * row.
    * <p>
    * This will return an array of drawable fragments, each fragment having
    * specified its start and end locations (both inclusive), relative to the
    * 'length' parameter.
    * 
    * @param    y
    *           The row location where to start looking for drawable fragments,
    *           horizontally.
    * @param    x
    *           The column location where to start looking for drawable 
    *           fragments, horizontally.
    * @param    length
    *           The length of the line to look into.
    *           
    * @return   A list of drawable fragments, where {@link Point#x} represents 
    *           the start row and {@link Point#y} represents the end row,
    *           inclusive.
    */
   @SuppressWarnings("unchecked")
   List<Point> getVerticalFragments(int y, int x, int length)
   {
      if (y < 0 || x < 0 || y >= rows)
      {
         return Collections.emptyList();
      }

      List<Point> frags = null;
      
      int row = y;
      int end = Math.min(rows, row + length);
      int col = x;

      if (outerRectangle == null    || 
          (col >= outerRectangle.left()  && 
           col <= outerRectangle.right() &&
           !(row > outerRectangle.bottom() ||
             end < outerRectangle.top())))
      {
         // search only if the line might intersect the clipped area
         do
         {
            while (row < end && !canDraw(row, col))
            {
               row = row + BASE_UNIT;
            }
   
            int start = row;
            while (row < end && canDraw(row, col))
            {
               row = row + BASE_UNIT;
            }
            
            if (start != row)
            {
               if (frags == null)
               {
                  frags = new ArrayList<Point>(5);
               }
               frags.add(new Point(start - y, row - y - BASE_UNIT));
            }
         }
         while (row < end);
      }

      return frags == null ? Collections.EMPTY_LIST : frags;
   }

   /**
    * Get the drawable vertical fragments, to the down side of the specified
    * row.
    * <p>
    * This will return an array of drawable fragments, each fragment having
    * specified its start and end locations (both inclusive), relative to the
    * 'length' parameter.
    * 
    * @param    y
    *           The row location where to start looking for drawable fragments,
    *           horizontally.
    * @param    length
    *           The length of the line to look into.
    *           
    * @return   A list of drawable fragments, where {@link Point#x} represents 
    *           the start row and {@link Point#y} represents the end row,
    *           inclusive.
    */
   List<Point> getVerticalFragmentsAnyColumn(int y, int length)
   {
      if (y < 0 || y >= rows)
      {
         return Collections.emptyList();
      }
      
      List<Point> frags = new ArrayList<Point>();

      int row = y;
      int end = Math.min(rows, row + length);

      if (outerRectangle == null || !(row > outerRectangle.bottom() || end < outerRectangle.top()))
      {
         // search only if the line might intersect the clipped area
         Set<Point> rows = new TreeSet<>(new Comparator<Point>()
         {
            @Override
            public int compare(Point p1, Point p2)
            {
               return p1.x - p2.x;
            }
         });

         for (Rectangle r : bitmap)
         {
            rows.add(new Point(r.top(), r.bottom()));
         }
         
         Iterator<Point> itr = rows.iterator();
         if (itr.hasNext())
         {
            Point start = itr.next();
            do
            {
               int startRow = start.x;
               int endRow = start.y;
               
               // find the overlapping rows
               while (itr.hasNext())
               {
                  Point p = itr.next();

                  if (p.x - BASE_UNIT > start.y)
                  {
                     // these are not overlapping nor contiguous, terminate, but select new start
                     start = p;
                     break;
                  }
                  
                  if (p.y > start.y)
                  {
                     // overlapping, move downwards
                     endRow = p.y;
                     
                     // select new start
                     start = p;
                  }
               }
               
               frags.add(new Point(startRow - y, endRow - y));
            }
            while (itr.hasNext());
         }
      }
      return frags;
   }

   /**
    * Get number of bitmap rows.
    * 
    * @return  number of bitmap rows.
    */
   public int rows()
   {
      return rows;
   }

   /**
    * Get number of bitmap columns.
    * 
    * @return  number of bitmap columns.
    */
   public int cols()
   {
      return cols;
   }
}
